import unittest

from mock import Mock, patch
from suds import WebFault

from ..exceptions import ConnectError
from ..client import InteractClient


class InteractClientTests(unittest.TestCase):
    """ Test InteractClient """

    def setUp(self):
        self.client = Mock()
        self.configuration = {
            'username': 'username',
            'password': 'password',
            'pod': 'pod',
            'client': self.client,
        }
        self.interact = InteractClient(**self.configuration)

    @patch.object(InteractClient, 'WSDLS', {'pod': 'pod_wsdl'})
    def test_wsdl_property_returns_correct_value(self):
        self.assertEqual(self.interact.wsdl, 'pod_wsdl')

    @patch.object(InteractClient, 'ENDPOINTS', {'pod': 'pod_endpoint'})
    def test_endpoint_property_returns_correct_value(self):
        self.assertEqual(self.interact.endpoint, 'pod_endpoint')

    @patch.object(InteractClient, 'connect', Mock())
    def test_entering_context_calls_connect(self):
        self.assertFalse(self.interact.connect.called)
        with self.interact:
            self.assertTrue(self.interact.connect.called)

    @patch.object(InteractClient, 'disconnect', Mock())
    def test_leaving_context_calls_disconnect(self):
        with self.interact:
            self.assertFalse(self.interact.disconnect.called)
        self.assertTrue(self.interact.disconnect.called)

    @patch.object(InteractClient, 'login', Mock())
    def test_connect_method_calls_login(self):
        self.interact.connect()
        self.assertTrue(self.interact.login.called)

    @patch.object(InteractClient, 'login')
    def test_connect_method_raises_connect_error_on_account_fault(self, login):
        login.side_effect = WebFault(Mock(), Mock())
        with self.assertRaises(ConnectError):
            self.interact.connect()

    @patch.object(InteractClient, 'login')
    def test_connect_method_raises_connect_error_on_unknown_error(self, login):
        fault = Mock()
        del fault.detail.AccountFault
        login.side_effect = WebFault(fault, Mock())
        with self.assertRaises(ConnectError):
            self.interact.connect()

    @patch.object(InteractClient, 'login', Mock(return_value=Mock(sessionId=1)))
    def test_connect_method_returns_true_on_success(self):
        self.assertTrue(self.interact.connect())

    def test_connect_method_sets_soapheaders(self):
        soapheaders = Mock()
        self.interact.client.factory.create.return_value = soapheaders
        self.interact.connect()
        self.interact.client.set_options.assert_called_once_with(soapheaders=soapheaders)

    @patch.object(InteractClient, 'logout', Mock(return_value=True))
    def test_disconnect_method_returns_true_on_success(self):
        self.assertTrue(self.interact.disconnect())

    @patch.object(InteractClient, 'logout', Mock(return_value=False))
    def test_disconnect_method_returns_false_on_failure(self):
        self.assertFalse(self.interact.disconnect())

    def test_disconnect_method_unsets_soapheaders(self):
        self.interact.disconnect()
        self.interact.client.set_options.assert_called_once_with(soapheaders=())



import unittest

from mock import Mock

from ..types import InteractType


class InteractTypeTests(unittest.TestCase):
    def setUp(self):
        self.type = InteractType(foo='bar')
        self.client = Mock()

    def test_soap_attribute_method_sets_attribute(self):
        self.type.soap_attribute('red', True)
        self.assertTrue(self.type.red)

    def test_soap_attribute_method_registers_attribute(self):
        self.type.soap_attribute('blue', True)
        self.assertIn('blue', self.type._attributes)

    def test_soap_name_property_returns_class_name(self):
        self.assertEqual(self.type.soap_name, self.type.__class__.__name__)

    def test_instance_provides_attributes_through_dictionary_lookup(self):
        self.assertEqual(self.type.foo, self.type['foo'])

    def test_get_soap_object_method_returns_client_factory_type(self):
        soap_object = self.client.factory.create.return_value = Mock()
        self.assertEqual(self.type.get_soap_object(self.client), soap_object)

    def test_get_soap_object_method_returns_object_with_correct_attributes_set(self):
        self.type.soap_attribute('red_fish', True)
        soap_object = self.type.get_soap_object(self.client)
        self.assertTrue(all(
            [hasattr(soap_object, attr) for attr in ['redFish', 'foo']]))



class InteractType(object):

    """ InteractType class

    Provides base interact type functionality. Interact types should register their WSDL defined
    attributes via the soap_attribute method. This allows interact types to provide their own
    soap friendly objects for use with the suds client used by the InteractClient.

    Interact type attributes can be accessed via dictionary lookup, for example:

        >>> InteractType(foo=1)
        >>> InteractType['foo'] == InteractType.foo
        ... True
    """

    def __init__(self, *args, **kwargs):
        self._attributes = set()
        self.set_attributes(*args, **kwargs)

    def __getitem__(self, name):
        return getattr(self, name)

    @property
    def soap_name(self):
        """ Provide the WSDL defined name for this class. """
        return self.__class__.__name__

    def soap_attribute(self, name, value):
        """ Marks an attribute as being a part of the data defined by the soap datatype"""
        setattr(self, name, value)
        self._attributes.add(name)

    def get_soap_object(self, client):
        """ Create and return a soap service type defined for this instance """
        def to_soap_attribute(attr):
            words = attr.split('_')
            words = words[:1] + [word.capitalize() for word in words[1:]]
            return ''.join(words)

        soap_object = client.factory.create(self.soap_name)
        for attr in self._attributes:
            value = getattr(self, attr)
            setattr(soap_object, to_soap_attribute(attr), value)

        return soap_object

    def set_attributes(self, *args, **kwargs):
        for name, value in list(kwargs.items()):
            self.soap_attribute(name, value)


class InteractObject(InteractType):

    """ Responsys InteractObject Type """

    def set_attributes(self, folder_name, object_name):
        self.soap_attribute('folder_name', folder_name)
        self.soap_attribute('object_name', object_name)


class ListMergeRule(InteractType):

    """ ListMergeRule

    Constructor accepts overrides for the following defaults:

        'insert_on_no_match': True,
        'update_on_match': 'REPLACE_ALL',
        'match_column_name_1': 'Customer_Id_',
        'match_column_name_2': None,
        'match_column_name_3': None,
        'match_operator': 'NONE',
        'optin_value': 'I',
        'optout_value': 'O',
        'html_value': 'H',
        'text_value': 'T',
        'reject_record_if_channel_empty': 'E',
        'default_permission_status': 'OPTIN',

    See Responsys API documentation for more information on what values are available and what
    these options do.
    """

    DEFAULTS = {
        'insert_on_no_match': True,
        'update_on_match': 'REPLACE_ALL',
        'match_column_name_1': 'Customer_Id_',
        'match_column_name_2': None,
        'match_column_name_3': None,
        'match_operator': 'NONE',
        'optin_value': 'I',
        'optout_value': 'O',
        'html_value': 'H',
        'text_value': 'T',
        'reject_record_if_channel_empty': 'E',
        'default_permission_status': 'OPTIN',
    }

    def set_attributes(self, **overrides):
        options = self.DEFAULTS.copy()
        options.update(overrides)

        self.soap_attribute('insert_on_no_match', options['insert_on_no_match'])
        self.soap_attribute('update_on_match', options['update_on_match'])
        self.soap_attribute('match_column_name_1', options['match_column_name_1'])
        self.soap_attribute('match_column_name_2', options['match_column_name_2'])
        self.soap_attribute('match_column_name_3', options['match_column_name_3'])
        self.soap_attribute('match_operator', options['match_operator'])
        self.soap_attribute('optin_value', options['optin_value'])
        self.soap_attribute('optout_value', options['optout_value'])
        self.soap_attribute('html_value', options['html_value'])
        self.soap_attribute('text_value', options['text_value'])
        self.soap_attribute(
            'reject_record_if_channel_empty', options['reject_record_if_channel_empty'])
        self.soap_attribute('default_permission_status', options['default_permission_status'])


class RecordData(InteractType):

    """ Responsys RecordData Type

    Responsys type representing a mapping of field names to values. Inits with either a list of
    dict-like objects, or a RecordData type from the soap service. Can be iterated over and has
    a length.
    """

    def set_attributes(self, record_data):
        if getattr(record_data, 'fieldNames', None):
            # Handle RecordData Type
            field_names = record_data.fieldNames
            records = [Record(r) for r in record_data.records]
        else:
            # Handle list of dictionaries
            assert len(record_data), "Record list length must be non-zero"
            field_names = list(record_data[0].keys())
            records = [Record(r.values()) for r in record_data]

        self.soap_attribute('field_names', field_names)
        self.soap_attribute('records', records)

    def __iter__(self):
        for record in self.records:
            yield dict(zip(self.field_names, record.field_values))

    def __len__(self):
        return len(self.records)

    def get_soap_object(self, client):
        """ Override default get_soap_object behavior to account for child Record types """
        record_data = super().get_soap_object(client)
        record_data.records = [r.get_soap_object(client) for r in record_data.records]
        return record_data


class Record(InteractType):

    """ Responsys Record Type

    A record is a series of values. Can be iterated over and has a length.
    """

    def set_attributes(self, record):
        if getattr(record, 'fieldValues', None):
            # Handle API Record object
            field_values = record.fieldValues
        else:
            # Handle list of values
            field_values = list(record)

        self.soap_attribute('field_values', field_values)

    def __iter__(self):
        return (v for v in self.field_values)

    def __len__(self):
        return len(self.field_values)


class DeleteResult(InteractType):
    """ Responsys DeleteResult Type """
    def set_attributes(self, delete_result):
        self.soap_attribute('error_message', delete_result.errorMessage)
        self.soap_attribute('success', delete_result.success)
        self.soap_attribute('exception_code', delete_result.exceptionCode)
        self.soap_attribute('id', delete_result.id)


class LoginResult(InteractType):
    def set_attributes(self, login_result):
        self.soap_attribute('session_id', login_result.sessionId)


class MergeResult(InteractType):
    def set_attributes(self, merge_result):
        self.soap_attribute('insert_count', merge_result.insertCount)
        self.soap_attribute('update_count', merge_result.updateCount)
        self.soap_attribute('rejected_count', merge_result.rejectedCount)
        self.soap_attribute('total_count', merge_result.totalCount)
        self.soap_attribute('error_message', merge_result.errorMessage)


class RecipientResult(InteractType):
    def set_attributes(self, recipient_result):
        self.soap_attribute('recipient_id', recipient_result.recipientId)
        self.soap_attribute('error_message', recipient_result.errorMessage)


class ServerAuthResult(InteractType):
    def set_attributes(self, server_auth_result):
        self.soap_attribute('auth_session_id', server_auth_result.authSessionId)
        self.soap_attribute(
            'encrypted_client_challenge', server_auth_result.encryptedClientChallenge)
        self.soap_attribute('server_challenge', server_auth_result.serverChallenge)



import logging

from suds.client import Client
from suds import WebFault

from .exceptions import ConnectError
from .types import (
    RecordData, RecipientResult, MergeResult, DeleteResult, LoginResult, ServerAuthResult)

log = logging.getLogger(__name__)


class InteractClient(object):

    """ Interact Client Class

    Provides access to the methods defined by the Responsys Interact API. Example setup:

        >>> client = InteractClient(username, password, pod)
        >>> client.connect()
        >>> client.merge_list_members(interact_object, records, merge_rules)
        >>> client.disconnect()

    Using the client class as a context manager will automatically connect using the credentials
    provided, and disconnect upon context exit:

        >>> with InteractClient(username, password, pod) as client:
        ...     client.merge_list_members(interact_object, records, merge_rules)

    Since responsys limits the number of active sessions per account, this can help ensure you
    don't leave unused connections open.
    """

    WSDLS = {
        '2': 'https://ws2.responsys.net/webservices/wsdl/ResponsysWS_Level1.wsdl',
        '5': 'https://ws5.responsys.net/webservices/wsdl/ResponsysWS_Level1.wsdl',
        'rtm4': 'https://rtm4.responsys.net/tmws/services/TriggeredMessageWS?wsdl',
        'rtm4b': 'https://rtm4b.responsys.net/tmws/services/TriggeredMessageWS?wsdl',
    }

    ENDPOINTS = {
        '2': 'https://ws2.responsys.net/webservices/services/ResponsysWSService',
        '5': 'https://ws5.responsys.net/webservices/services/ResponsysWSService',
        'rtm4': 'http://rtm4.responsys.net:80/tmws/services/TriggeredMessageWS',
        'rtm4b': 'http://rtm4b.responsys.net:80/tmws/services/TriggeredMessageWS',
    }

    @property
    def wsdl(self):
        return self.WSDLS[self.pod]

    @property
    def endpoint(self):
        return self.ENDPOINTS[self.pod]

    def __init__(self, username, password, pod, client=None):
        self.username = username
        self.password = password
        self.pod = pod
        self.client = client or Client(self.wsdl, location=self.endpoint)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type_, value, traceback):
        self.disconnect()

    def connect(self):
        """ Connects to the Responsys soap service

        Uses the credentials passed to the client init to login and setup the session id returned.
        Returns True on successful connection, otherwise False.
        """
        try:
            login_result = self.login(self.username, self.password)
        except WebFault as e:
            account_fault = getattr(e.fault.detail, 'AccountFault', None)
            if account_fault:
                log.error('Login failed, invalid username or password')
            else:
                log.error('Login failed, unknown error', exc_info=True)
            raise ConnectError("Failed to connect to soap service")

        self.__set_session(login_result.session_id)
        return True

    def disconnect(self):
        """ Disconnects from the Responsys soap service

        Calls the service logout method and destroys the client's session information. Returns
        True on success, False otherwise.
        """
        if self.logout():
            self.__unset_session()
            return True
        return False

    # Session Management Methods
    def login(self, username, password):
        """ Responsys.login soap call

        Accepts username and password for authentication, returns a LoginResult object.
        """
        return LoginResult(self.client.service.login(username, password))

    def logout(self):
        """ Responsys.logout soap call

        Returns True on success, False otherwise.
        """
        return self.client.service.logout()

    def login_with_certificate(self, encrypted_server_challenge):
        """ Responsys.loginWithCertificate soap call

        Accepts encrypted_server_challenge for login. Returns LoginResult.
        """
        return LoginResult(self.client.service.loginWithCertificate(encrypted_server_challenge))

    def authenticate_server(self, username, client_challenge):
        """ Responsys.authenticateServer soap call

        Accepts username and client_challenge to authenciate. Returns ServerAuthResult.
        """
        return ServerAuthResult(self.client.service.authenticateServer(username, client_challenge))

    def __set_session(self, session_id):
        """ Set appropriate session header on suds client """
        session_header = self.client.factory.create('SessionHeader')
        session_header.sessionId = session_id
        self.client.set_options(soapheaders=session_header)

    def __unset_session(self):
        """ Remove session header from current suds client """
        self.client.set_options(soapheaders=())

    # List Management Methods
    def merge_list_members(self, list_, record_data, merge_rule):
        """ Responsys.mergeListMembers call

        Accepts:
            InteractObject list_
            RecordData record_data
            ListMergeRule merge_rule

        Returns a MergeResult
        """
        list_ = list_.get_soap_object(self.client)
        record_data = record_data.get_soap_object(self.client)
        merge_rule = merge_rule.get_soap_object(self.client)
        return MergeResult(self.client.service.mergeListMembers(list_, record_data, merge_rule))

    def merge_list_members_RIID(self, list_, record_data, merge_rule):
        """ Responsys.mergeListMembersRIID call

        Accepts:
            InteractObject list_
            RecordData record_data
            ListMergeRule merge_rule

        Returns a RecipientResult
        """
        list_ = list_.get_soap_object(self.client)
        result = self.client.service.mergeListMembersRIID(list_, record_data, merge_rule)
        return RecipientResult(result.recipientResult)

    def delete_list_members(self, list_, query_column, ids_to_delete):
        """ Responsys.deleteListMembers call

        Accepts:
            InteractObject list_
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list ids_to_delete

        Returns a list of DeleteResult instances
        """
        list_ = list_.get_soap_object(self.client)
        result = self.client.service.deleteListMembers(list_, query_column, ids_to_delete)
        if hasattr(result, '__iter__'):
            return [DeleteResult(result) for delete_result in result]
        return [DeleteResult(result)]

    def retrieve_list_members(self, list_, query_column, field_list, ids_to_retrieve):
        """ Responsys.retrieveListMembers call

        Accepts:
            InteractObject list_
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list field_list
            list ids_to_retrieve

        Returns a RecordData instance
        """
        list_ = list_.get_soap_object(self.client)
        result = self.client.service.retrieveListMembers(
            list_, query_column, field_list, ids_to_retrieve)
        return RecordData(result.recordData)

    # Table Management Methods
    def delete_table_records(self, table, query_column, ids_to_delete):
        """ Responsys.deleteTableRecords call

        Accepts:
            InteractObject table
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list ids_to_delete

        Returns a list of DeleteResult instances
        """
        table = table.get_soap_object()
        result = self.client.service.deleteTableRecords(table, query_column, ids_to_delete)
        if hasattr(result, '__iter__'):
            return [DeleteResult(result) for delete_result in result]
        return [DeleteResult(result)]

    def merge_table_records_with_pk(self, table, record_data, insert_on_no_match, update_on_match):
        """ Responsys.mergeTableRecordsWithPK call

        Accepts:
            InteractObject table
            RecordData record_data
            string insert_on_no_match
            string update_on_match

        Returns a RecipientResult
        """
        table = table.get_soap_object()
        record_data = record_data.get_soap_object()
        return RecipientResult(self.client.service.mergeTableRecordsWithPK(
            table, record_data, insert_on_no_match, update_on_match))

    def merge_into_profile_extension(self, profile_extension, record_data, match_column,
                                     insert_on_no_match, update_on_match):
        """ Responsys.mergeIntoProfileExtension call

        Accepts:
            InteractObject profile_extension
            RecordData record_data
            string match_column
            string insert_on_no_match
            string update_on_match

        Returns a RecipientResult
        """
        profile_extension = profile_extension.get_soap_object()
        record_data = record_data.get_soap_object()
        return RecipientResult(self.client.service.mergeIntoProfileExtension(
            profile_extension, record_data, match_column, insert_on_no_match, update_on_match))

    def retrieve_table_recods(self, table, query_column, field_list, ids_to_retrieve):
        """ Responsys.mergeIntoProfileExtension call

        Accepts:
            InteractObject table
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list field_list
            list ids_to_retrieve

        Returns a RecordData
        """
        table = table.get_soap_object()
        return RecordData(self.client.service.retrieveTableRecords(
            table, query_column, field_list, ids_to_retrieve))

    # TODO: Implement
    #
    # Content Management Methods
    # Folder Management Methods,
    # Campagin Management Methods



import unittest
import logging

from ..client import InteractClient
from ..types import InteractObject, ListMergeRule, RecordData


logging.disable(logging.CRITICAL)

class ResponsysIntegrationTest(unittest.TestCase):
    "Responsys Integration InteractClient"

    def setUp(self):
        self.MASTER_FOLDER = 'UDEMY_TEST_FOLDER'
        self.MCL_OBJECT = 'CONTACTS_LIST_TEST'
        self.username = 'api@udemy'
        self.password = 'Murfyzf8TG8cuNY'
        self.pod = '5'
        self.client = InteractClient(self.username, self.password, self.pod)

    def test_connects_and_does_stuff(self):
        with self.client as client:
            list_ = InteractObject(self.MASTER_FOLDER, self.MCL_OBJECT)
            members = client.retrieve_list_members(
                list_,
                'CUSTOMER_ID',
                ['First_Name', 'Last_Name', 'Email_Address_'],
                list(range(1, 200))
            )
            self.assertTrue(len(members))
            record_data = [{
                 'Customer_Id_': 90000,
                 'First_Name': 'Foo',
                 'Last_Name': 'Bar',
                 'Email_Address_': 'email@email.com',
                 'Email_Permission_Status_': 'I',
                 'Registration_Date': '01/01/01',
                 'OPTOUT_LINK': '',
                 'COUNTRY_': '',
                 'Lang_Locale': '',
                 'Date_Of_Birth': '',
                 'Gender': '',
                 'Active_Status': '',
                 'Course_Category_1': '',
                 'Course_Category_2': '',
                 'Course_Category_3': '',
                 'Course_Category_4': '',
                 'Course_Category_5': '',
                 'Course_Category_6': '',
                 'Course_Category_7': '',
                 'Course_Category_8': '',
                 'Course_Category_9': '',
                 'Course_Category_10': '',
                 'Course_Category_11': '',
                 'Course_Category_12': '',
                 'Course_Category_13': '',
                 'Course_Category_14': '',
                 'Course_Category_15': '',
                 'Last_Login': '',
                 'Number_Of_Logins': '',
                 'Number_Of_Published_Courses': '',
                 'Number_Of_Published_HQ_Courses': '',
                 'Number_Of_Free_Courses': '',
                 'Number_Of_Paid_Courses': '',
                 'Number_Of_Courses_Created': '',
                 'FB_User_Id': '',
                 'Signup_Source': '',
                 'Is_Premium_Instructor': '',
                 'Is_Affiliate': '',
                 'POSTAL_STREET_1_': '',
                 'CITY_': '',
                 'STATE_': '',
                 'Postal_Code_': '',
                 'User_Locale': 'en_US',
                 'Platform': '',
                 'SOURCE_ORGANIZATION_ID': '',
                 'LAST_PURCHASE_DATE': '',
                 'LAST_PAID_PURCHASE_DATE': '',
                 'TIMEZONE': '',
                 'UTC_DIFF': '',
            }]
            record_data = RecordData(record_data)
            merge_response = client.merge_list_members(list_, record_data, ListMergeRule())
            client.delete_list_members(list_, 'CUSTOMER_ID', [90000])
            self.fail()



from setuptools import setup, find_packages
from os import path

import responsys

project_dir = path.abspath(path.dirname(__file__))
with open(path.join(project_dir, 'README.md')) as f:
    long_description = f.read()

install_requires = (
    'suds-jurko'
)
tests_require = (
    'coverage',
    'mock',
    'nose',
    'pinocchio',
)

setup(
    name=responsys.__name__,
    keywords=responsys.__keywords__,
    version=responsys.__version__,
    author='Jared Lang',
    author_email='kaptainlange@gmail.com',
    description='Python client library for the Responsys Interact API',
    long_description=long_description,
    packages=find_packages(),
    license='LICENSE',
    install_requires=install_requires,
    setup_requires=tests_require,
    tests_require=tests_require,
    test_suite='nose.collector',
)
